/*
 * ----- Input Devices Abstraction Layer: ----- 
 */

boolean confirm(prog_char const msg[]) {
  lcdPrintProg2(msg, yesNoMsg);
  waitForKeyUp();
  yellowOn();
    
  int8_t key = -1;
  while (key != KEY_LEFT && key != KEY_RIGHT) {
    key = getKeyPressed();
  }      
  waitForKeyUp();
  return (key == KEY_RIGHT);
}

// --------------------------------

int8_t enterKey() {
  yellowOn();
  int8_t key = -1;
  while ((key = getKeyPressed()) < 0);
  yellowOff();
  return key;
}

// --------------------------------

int8_t getKeyPressed() {
  uint8_t cnt = 0;
  int8_t  key = getKey();
  if (key >= 0) {
    while (getKey() >= 0) {
      if (cnt < 255) {
        cnt++;
      }
      if (cnt > 20) {
        flashRed();
      }
      delay50();
    }
  }
  keyPressedTime = cnt;
  return key;
} 

// --------------------------------

void waitForKeyUp() {
  while (getKey() >= 0);

  yellowOff();
  lcdNoCursor();
  lcdNoBlink();
}

// --------------------------------

#ifndef HWKEYS
//static int    adc_key_val[] = { 30, 175, 360, 535, 760 }; // thresholds of the cascading resistors
static int    adc_key_val[] = { 64, 220, 390, 560, 860 }; // thresholds of the cascading resistors
static int8_t key_mapping[] = { KEY_RIGHT, KEY_UP, KEY_DOWN, KEY_LEFT, KEY_SELECT, };

int8_t getRawKey() {
  #if defined(KEY_PIN)
  rawKey = analogRead(KEY_PIN);   // read the value from the sensor
  for (uint8_t k = 0; k < NUM_KEYS; k++) {
    if (rawKey < adc_key_val[k]) {
      return key_mapping[k];
    }
  }
  #endif
  return -1;
}

#else //HWKEYS

int8_t getRawKey() {
  if      (digitalRead(HW_RIGHT) ==0) {return KEY_RIGHT; }
  else if (digitalRead(HW_UP)    ==0) {return KEY_UP;    }
  else if (digitalRead(HW_DOWN)  ==0) {return KEY_DOWN;  }
  else if (digitalRead(HW_LEFT)  ==0) {return KEY_LEFT;  }
  else if (digitalRead(HW_SELECT)==0) {return KEY_SELECT;}
}

#endif // HWKEYS

// --------------------------------

#ifdef JOYSTICK

int8_t getRawJoystick() {
  #ifdef JS_SELECT_PIN
  int sel = analogRead(JS_SELECT_PIN);
  if (sel < 100) {
    return KEY_SELECT;
  }
  #endif

  jsHorAct  = analogRead(JS_HOR_PIN);
  jsVertAct = analogRead(JS_VERT_PIN);
  
  if (jsHorAct < 1000 || jsVertAct < 1000) { // is joystick present? (int. pullups are on)
    int hor  = (jsHorAct  - jsHorCtr ) * JS_HOR_SIGN;
    int vert = (jsVertAct - jsVertCtr) * JS_VERT_SIGN;

    if (vert < JS_DEADBAND && vert > -JS_DEADBAND) {
      if      (hor < -5 * JS_DEADBAND)  {return KEY_RIGHT;}
      else if (hor >  5 * JS_DEADBAND)  {return KEY_LEFT; }
    }
    else if (hor < JS_DEADBAND && hor > -JS_DEADBAND) {
      if      (vert < -5 * JS_DEADBAND) {return KEY_UP;   }
      else if (vert >  5 * JS_DEADBAND) {return KEY_DOWN; }
    }
  }
  return -1;
}

#endif // JOYSTICK

// --------------------------------

int8_t getKey() {
  int8_t key1 = getRawKey();      // read the value from the sensor
  if (key1 >= 0) {                // is key pressed?
    delay50();                    // debounce time
    int8_t key2 = getRawKey();    // read the sensor value again
    if (key1 == key2) {
      return key1;
    }
  }

#ifdef JOYSTICK
   key1 = getRawJoystick();       // read the value from the sensor
  if (key1 >= 0) {                // is key pressed?
    delay50();                    // debounce time
    int8_t key2 = getRawJoystick(); // read the sensor value again
    if (key1 == key2) {
      return key1;
    }
  }
#endif
  
  return -1;
}
