/*
 * ---- Input Methods: -----
 */

void enterPid(int8_t index) {
  static uint8_t col [] = { 4, 10, 15, };
//  static uint8_t maxv[] = { 250, 250, 40, };
  
  int8_t idx = index - 1;
  if (idx < 0) {idx = 0;}
 
  uint8_t val[3];
  val[0] = P8[idx]; 
  val[1] = I8[idx]; 
  val[2] = D8[idx];
  
  int8_t local    =  0; // local state (-1..4)
  int8_t lastKey  = -1;
  int    count    =  0;
  int    repeat   =  0;
  int8_t maxState =  3;
  if      (index == 8)              {maxState = 1;}    // Mag   - only P
  else if (index == 7 && vers < 20) {maxState = 2;}    // Level - only PI for v1.8 and v1.9
  
  lcdBlink();
  while (local >= 0 && local < maxState) {
    lcdSetCursor(col[local] + 16);

    int8_t key = getKey();
    if (key < 0) {
      lastKey = key;
      repeat  = 0;
    }
    else if (key == KEY_UP || key == KEY_DOWN || key != lastKey) { // no repeat for LEFT, RIGHT, SEL
      if (key != lastKey) {
        repeat = 0;
      }
      lastKey = key;
      if (key == KEY_LEFT) {
        local--;
      }
      else if (key == KEY_SELECT || (key == KEY_RIGHT && local == maxState-1)) {
        uint8_t pp = val[0];
        uint8_t ii = val[1];
        uint8_t dd = val[2];

        if (index == 0) {
          P8[1] = pp;
          I8[1] = ii;
          D8[1] = dd;
        }
        P8[idx] = pp;
        I8[idx] = ii;
        D8[idx] = dd;
        
        serialWrite();
        local = -1;
      }
      else if (key == KEY_RIGHT) {
        local++; 
      }
      else {
        if (repeat > 3 || (count % 8) == 0) {
          uint8_t tmp = val[local];
          if (key == KEY_DECR) {
            if (tmp > 0) {
              tmp--;
            }
          }
          else {
            if (tmp < 250) { // maxv[local]
              tmp++;
            }
          }
          val[local] = tmp;
          if      (local == 0) { format      ( 0 + 16,  tmp, 4, 2); }
          else if (local == 1) { format      ( 5 + 16,  tmp, 5, 4); }    
          else if (local == 2) { formatSigned(11 + 16, -tmp); } 
          repeat++;
        }
      }
    }
    count++;
//    delay50();
  }
  waitForKeyUp();
}

// --------------------------------

void enterRates(int8_t index) {
  uint16_t val[3], v0, v1, v2;

  v0 = rcRate8;
  v1 = rcExpo8;
  v2 = powerTrigger;

  if (index == 1) {
    v0 = rollPitchRate;
    v1 = yawRate;
    v2 = dynThrPID;
  }

  val[0] = v0;
  val[1] = v1;
  val[2] = v2;

  int8_t minState = (index == 2)? 2 : 0;
  int8_t maxState = (index == 0)? 2 : 3;
  int8_t local    =  minState;  // local state (-1..4)
  int8_t lastKey  = -1;
  int    count    =  0;
  int    repeat   =  0;

  lcdBlink();
  while (local >= minState && local < maxState) {
    uint8_t col = local * 6 + 4 + 16;
    if (col > 31) {col = 31;}
    lcdSetCursor(col);
    
    int8_t key = getKey();
    if (key < 0) {
      lastKey = -1;
      repeat  = 0;
    }
    else if (key == KEY_UP || key == KEY_DOWN || key != lastKey) { // no repeat for LEFT, RIGHT, SEL
      if (key != lastKey) {
        repeat = 0;
      }
      lastKey = key;
      if (key == KEY_LEFT) {
        local--;
      }
      else if (key == KEY_SELECT || (key == KEY_RIGHT && local == maxState-1)) {
        v0 = val[0];
        v1 = val[1];
        v2 = val[2];

        if (index == 0) {
          rcRate8       = v0;
          rcExpo8       = v1;
        }
        else if (index == 1) {
          rollPitchRate = v0;
          yawRate       = v1;
          dynThrPID     = v2;
        }
        else {
          powerTrigger  = v2;
        }

        serialWrite();
        local = -1;
      }
      else if (key == KEY_RIGHT) {
        local++; 
      }
      else {
        if (repeat > 3 || (count % 8) == 0) {
          uint8_t  incr = (index == 2)? 10 : 1;
          uint16_t tmp  = val[local];
          uint16_t maxv = 100;
          if (index == 2) { 
            maxv = 9990;
          }
          else if (index == 0 && local != 1) {
            maxv = 250;
          }

          if (key == KEY_DECR) {
            if (tmp > 0) {
              tmp -= incr;
            }
          }
          else {
            if (tmp < maxv) {
              tmp += incr;
            }
          }

          val[local] = tmp;
          if (index == 0 && local == 0) {
            tmp *= 2; // RC Rate
          }
          format(local * 6 + 16, tmp);
          repeat++;
        }
      }
    }
    count++;
//    delay50();
  }
  waitForKeyUp();
}

// --------------------------------

boolean enterName(char name[]) {
  int8_t local   =  0; // local state (-1..8)
  int8_t lastKey = -1;
  int    count   =  0;
  int    repeat  =  0;
  
  for (uint8_t i = 0; i < 8; i++) {
    if (name[i] == '-') {
      name[i] = ' ';
    }
  }
  char nn = name[0];
  
  lcdCursor();
  lcdWrite    (4+16, nn);
  lcdSetCursor(4+16);
  
  while (local >= 0 && local < 8) {
    int8_t key = getKey();
    if (key < 0) {
      lastKey = -1;
      repeat  = 0;
    }
    else if (key == KEY_UP || key == KEY_DOWN || key != lastKey) { // no repeat for LEFT, RIGHT, SEL
      if (key != lastKey) {
        repeat = 0;
      }
      lastKey = key;
      boolean refresh = false;
      if (key == KEY_LEFT) {
        local--;
        refresh = (local >= 0);
        if (refresh) {
          nn = name[local];
        }
      }
      else if (key == KEY_RIGHT) { 
        local++; 
        refresh = (local < 8);
        if (refresh) { // handle RIGHT in third column as SELECT
          nn = name[local];
        }
      }
      else if (key == KEY_SELECT) {
        local = 8;
      }
      else {
        if (repeat > 3 || (count % 8) == 0) {
          if (key == KEY_DECR) {
            if (nn > 32) {
              nn--;
              refresh = true;
            }
          }
          else {
            if (nn < 125) {
              nn++;
              refresh = true;
            }
          }
          if (refresh) {
            name[local] = nn;
          }
        }
      }

      if (refresh) {
        lcdWrite    (local + 4 + 16, name[local]);
        lcdSetCursor(local + 4 + 16);
        repeat++;
      }
    }
    count++;
  }
  waitForKeyUp();
  
  return (local == 8);
}

// --------------------------------

void enterAuxState() {
  lcdCursor();
  
  int8_t   local = 0; // local state (-1..6)
  uint16_t old   = actMode[auxState];
  int8_t   key   = 0;

  while (local >= 0 && local < 12) {
    if (key >= 0) {
      printAuxState();

      uint8_t pos = 17 + local + (local / 3);
      lcdSetCursor(pos);
    }

    key = getKeyPressed();
    if      (key == KEY_LEFT)   {local--;}
    else if (key == KEY_RIGHT)  {local++;}
    else if (key == KEY_SELECT) {local = 12;}
    else if (key >= 0) {
      uint16_t mode     = actMode[auxState];
      mode             ^= auxMasks[local];
      actMode[auxState] = mode;
    }
//    delay50();
  }
  waitForKeyUp();

  if (local == 12) {
    serialWrite();
  }
  else {
    actMode[auxState] = old;
  }
}

// --------------------------------

void enterValue(int8_t index) {  
  bool showValue = (index < 0);
  if (showValue) {
    index = -index;
  }
  else {
    lcdWrite(':');
    lcdBlink();
  }
  
  uint16_t val = baudRate;
  if      (index == 1) { val = delayTime; lcdPrintProg1(delaySuff); }
#ifdef BACKLIGHT
  else if (index != 2) { val = backlight; }
#endif

  int    repeat  =  0;
  int    count   =  0;
  int8_t lastKey = -1;

  while (true) {
    if      (index == 1) { format( 5 + 16, val * 10); }
    else if (index == 2) { format( 5 + 16, baudRates[val]*12); lcdPrint("00"); }
    else                 { format( 5 + 16, val); }
    
    if (showValue) {
      return;
    }
    
    int8_t key = getKey();
    if (key < 0) {
      lastKey = -1;
      repeat  =  0;
    }
    else if (key == KEY_UP || key == KEY_DOWN || key != lastKey) { // no repeat for LEFT, RIGHT, SEL
      if (key != lastKey) {
        repeat = 0;
      }
      lastKey = key;
      if (key == KEY_UP || key == KEY_DOWN) {
        if (repeat > 3 || (count % 8) == 0) {
          int8_t  incr = 1;
          uint8_t maxV = 250;

          if (index == 2) {
            maxV = 8;
          }
          else { 
            incr = 5;
          }

          if (key == KEY_DECR) {
            if (val > 0) {
              val -= incr;
            } 
          }
          else {
            if (val < maxV) {
              val += incr;
            }
          }

#ifdef BACKLIGHT
          if (index == 3) {
            setBacklight(val);
          }
#endif
          repeat++;
        }
      }
      else {
        if (key != KEY_LEFT) {
          if      (index == 1) { delayTime = val; }
          else if (index == 2) { baudRate  = val; }
#ifdef BACKLIGHT
          else                 { backlight = val; }
#endif  
          storeParams();
        }
        waitForKeyUp();
        return; 
      }
    }
    count++;
//    delay50();
  }
}
