/*
 * ----- Serial Communication: ----- 
 */

void    serialFlush()             {Serial.flush();}
boolean serialAvailable()         {return Serial.available();}
void    serialBegin()             {Serial.begin(long(baudRates[baudRate])*1200);}  
uint8_t serialReadByte()          {Serial.read();}  
void    serialWrite(uint8_t ch)   {Serial.write(ch);}  

void serialize8   (int8_t   val)  {*bptr++ = val;}
void serialize16  (int16_t  val)  {*bptr++ = val; *bptr++ = val/256;}
void serialize16  (uint16_t val)  {*bptr++ = val; *bptr++ = val/256;}

void deserialize8 (int8_t&  val)  {val = *bptr++;}
void deserialize8 (uint8_t& val)  {val = *bptr++;}

void deserialize16(int16_t&  val) {uint8_t low = *bptr++; uint8_t hi = *bptr++; val = (int16_t)  (hi * 256 + low);}
void deserialize16(uint16_t& val) {uint8_t low = *bptr++; uint8_t hi = *bptr++; val = (uint16_t) (hi * 256 + low);}

void deserialize16(int16_t arg[]) {
  deserialize16(arg[0]); 
  if (state >= 0) {
    if (arg[1] > arg[0]) {arg[1] = arg[0];}
    if (arg[2] < arg[0]) {arg[2] = arg[0];}
  }
  else {
    arg[2] = arg[1] = arg[0];
  }
}

void deserialize16(int8_t len, int16_t arr[][3]) {for (int8_t i = 0; i < len; i++) deserialize16(arr[i]);}
void deserialize16(int8_t len, int16_t arr[]   ) {for (int8_t i = 0; i < len; i++) deserialize16(arr[i]);}

// --------------------------------

#ifndef TEST

uint8_t serialRead() {
  uint8_t cnt = 0;
  int16_t res = 0;
  if (!offline) {
    while (cnt <= 3 && (res = innerSerialRead()) == -1) {
      cnt++;
    }
  
    if (res < 0) {
#ifdef DUMP_BUFFER
      dumpBuffer();
#endif  
      lcdPrintProg1(haltMsg); 
      while (true) {
        blinkLed(RED_LED, 125);
      }
    }
    else if (res == 0) {
      lcdPrintProg1(cancelRead);
      blinkWarning();
      refreshDisplay();
      offline = true;
      vers    = 20;
    }
    else if (state != 35) { // no flashing in battery level mode
      flashGreen();
    }
  }
  else {
    flashRed();
  }
  return res;
}

int16_t innerSerialRead() {
  if (state < 0) {
    lcdPrintProg1(reading);
  }

  while (serialAvailable()) { // clear inbuffer
    serialReadByte();
  }

  char tag = 'M';
  
  uint16_t retry    = 30000; // 30000 * 2 ms = 60 sec timeout
  int16_t length    = 3;
  int16_t bytesRead = 0;
  boolean readFlag  = false;
  
  bptr = buffer;
  while (retry > 0 && bytesRead < length) {
    if (state < 0) {
      if (readFlag) {
        format(9 + 16, bytesRead, 3, 0);
      }
      int8_t key = getKey();
      switch (key) {
#ifdef EMULATE_TEXTSTAR
        case KEY_RIGHT:
          emulateLcd();  // no break!
#endif            
        case KEY_LEFT:
          // cancel loop
          return 0;
      }
    }

    readFlag = false;
    while (serialAvailable() && bytesRead < BUFFER_SIZE) {
      readFlag     = true;
      retry        = 250;
      int8_t  err  = -1;
      uint8_t last = serialReadByte();
      *bptr++      = last;
      bytesRead++;
      
      if (bytesRead == 1) {
        if (last != tag) {
          err = 1;  // wrong prefix
        }
      }
      else if (bytesRead == 2) {
        uint8_t vIndex = (last > 200)? last/100 : last-18;
        if (vIndex > sizeof(frameLength)) {
          err = 4; // wrong version
        }
        else {
          length = frameLength[vIndex];
        }
      }
      else if (bytesRead == length) {
        if (last != tag) {
          err = 2; // wrong suffix
        }
        else {
          deserialize();
          if (config >= NUM_CONFIGS) {
            err  = 3; // wrong config
            last = config;
          }
        }
      }
      if (err >= 0) {
        showFrameError(err, last);
        return -1;
      }
    }

    if ((retry % 500) == 0) {  // redo after 1 sec
      if (retry <= 29500) {
        flashRed();            // start flashing after 1 sec = 2 ms * 500
      }        
      if (bytesRead == 0) {
        serialWrite(tag);
        serialFlush();
      }
    }
    delay(2);
    retry--;
  }
  
  if (bytesRead == 0) {
    lcdPrintProg0(timeoutError);
    return -2; // stop immediately
  }
  if (bytesRead < length) {
    showFrameError(0, bytesRead);
    return -1; // signal error, retry
  }
  return bytesRead;
}

void showFrameError(uint8_t code, int16_t arg) {
  lcdPrintProg0(wrongMsg);
  lcdPrintProg (6, frameError[code]);
  format(13, arg, 3, 0);
  blinkWarning();
  refreshDisplay();
}

// --------------------------------

#else // TEST

uint8_t serialRead() {
  bptr = buffer;
  int8_t i;

  serialize8('M');                                     // 00: tag
  serialize8(201);                                     // 01: version

  for (i = 0; i < 3; i++) serRnd(accSmooth[i]);        // 02-07: acc
  for (i = 0; i < 3; i++) serRnd(gyroData [i]);        // 08-13: gyro
  for (i = 0; i < 3; i++) serRnd(magADC   [i]);        // 14-19: mag

  serialize16(333);                                    // 20-21: altitude
  serialize16(100);                                    // 22-23: heading

  for (i = 0; i < 8; i++) serRnd(servo       [i]);     // 24-39: servos
  for (i = 0; i < 8; i++) serRnd(motor       [i]);     // 40-55: motor
  for (i = 0; i < 8; i++) serRnd(rcHysteresis[i]);     // 56-71: receiver

  serialize8(31);                                      // 72: avail. sensors:   0-nunchukPres, 1-accPres, 2-baroPres, 3-magPres, 4-gpsPres
  serialize8(63);                                      // 73: sensor modes:     0-accMode, 1-baroMode, 2-magMode, 3-gpsModeHome, 4-gpsModeHold, 5-armed

  serialize16(5120);                                   // 74-75: cycle time
  serialize16(7);                                      // 76-77: i2c error coun
  
  for (i = 0; i < 2; i++) {
    serRnd(angle[i]);                                  // 78-81: roll and pitch angle
  }
  
  if (config == 0) {
    config = 1;
  }
  serialize8(config);                                  // 82: config / multitype (Tri, Quad, ...)

  for (i = 0; i < 8; i++) {
    serialize8(40); serialize8(30); serialize8(13);    // 83-106: P-I-D (roll, pitch, yaw, alt, vel, gps, level, mag)
  }
  
  serialize8(50);                                      // 107: rcRate (dual rate) [value is doubled]
  serialize8(66);                                      // 108: rcExpo
  serialize8(25);                                      // 109: roll/pitch rate
  serialize8(50);                                      // 110: yaw rate
  serialize8(75);                                      // 111: dyn. throttle PID
  
  for (i = 0; i < AUX_STATES; i++) {                   // 112-133: activation modes
    serialize16(auxMasks[i] | 0x80);
  }

  serialize16(4711);                                   // 134-135: gpsDistToHome
  serialize16(99);                                     // 136-137: gpsDirection
  serialize8(5);                                       // 138: gpsNumSat
  serialize8(14);                                      // 139: gpsFix
  serialize8(77);                                      // 140: gpsUpdate

  serialize16(3000);                                   // 141-142: powerMeterSum
  serialize16(40000);                                  // 143-144: powerTrigger
  serialize8(90 + random(30));                         // 145: vbat (battery level)

  serialize16(10);                                     // 146-153: 4 16bit debug values
  serialize16(100);                                  
  serialize16(1000);                                   
  serialize16(10000);                                  
  
  serialize8('M');                                     // 154 - tag
    
  deserialize();
  
  if (state != 35) {  // no flashing in battery level mode
    flashGreen();
  }
  delay(200);

  return (bptr - buffer);
}

void serRnd(int16_t val MM_DIM) {
  serialize16(val MM_ACT + random(21) - 10);
}
#endif // TEST

// --------------------------------

void deserialize() {
  bptr = buffer + 1;                           // 00: tag
  int8_t i;

  deserialize8(vers);                          // 01: version
  if (vers > 200) {
    vers /= 10;
  }

  deserialize16(3, accSmooth);                 // 02-07: acc
  deserialize16(3, gyroData );                 // 08-13: gyro
  deserialize16(3, magADC   );                 // 14-19: mag

  deserialize16(altitude);                     // 20-21: altitude
  deserialize16(heading);                      // 22-23: heading

  int8_t numServos = (vers < 20)? 4 : 8;

  deserialize16(numServos, servo       );      // 24-39: servos
  deserialize16(8,         motor       );      // 40-55: motor
  deserialize16(8,         rcHysteresis);      // 56-71: receiver

  int8_t sensorData;  
  deserialize8(sensorData);                    // 72: avail. sensors:   0-nunchukPres, 1-accPres, 2-baroPres, 3-magPres, 4-gpsPres
        
  accPres      = (sensorData & 3);
  baroPres     = (sensorData & 4);
  magPres      = (sensorData & 8);
  gpsPres      = (sensorData & 16);

  deserialize8(sensorData);                    // 73: sensor modes: 0-accMode, 1-baroMode, 2-magMode, 3-gpsModeHome, 4-gpsModeHold, 5-armed
  accMode      = (sensorData & 1);
  baroMode     = (sensorData & 2);
  magMode      = (sensorData & 4);
  gpsModeHome  = (sensorData & 8);
  gpsModeHold  = (sensorData & 16);
  armed        = (sensorData & 32);

  deserialize16(cycleTime);                    // 74-75: cycle time
  if (vers == 20) {
    deserialize16(i2c_errorCount);             // 76-77: i2c error count
  }
  
  deserialize16(2, angle);                     // 78-81: roll and pitch angle
  
  deserialize8(config);                        // 82: config / multitype (Tri, Quad, ...)

  if (vers < 20) {
    for (i = 0; i < 7; i++) {
      int8_t j = (i > 4)? i+1 : i;             // map old PID indexes to v2.0 indexes
      deserialize8(P8[j]);                     // (73-87): P-I-D (roll, pitch, yaw, alt, vel)
      if (i < 6) {
        deserialize8(I8[j]);
        if (i < 5) {
          deserialize8(D8[j]);
        }
      }
    }
  }
  else {
    for (i = 0; i < 8; i++) {
      deserialize8(P8[i]);                     // 83-106: P-I-D (roll, pitch, yaw, alt, vel, gps, lev, mag)
      deserialize8(I8[i]);
      deserialize8(D8[i]);
    }
  }
  
  deserialize8(rcRate8);                       // 107: rcRate (dual rate)
  deserialize8(rcExpo8);                       // 108: rcExpo
  deserialize8(rollPitchRate);                 // 109: roll/pitch rate
  deserialize8(yawRate);                       // 110: yaw rate
  deserialize8(dynThrPID);                     // 111: dyn. throttle PID
  
  if (vers < 20) {
    int actLen = (vers == 19)? 8 : 6;  
    for (i = 0; i < actLen; i++) {             // (96-103): activation modes (level, baro, mag, camStab, camTrig, arming)
      uint8_t tmp;
      deserialize8(tmp);
      actMode[i] = tmp;
    }
  }
  else {
    for (i = 0; i < AUX_STATES; i++) {         // 112-133: activation modes
      deserialize16(actMode[i]);
    }
  }
    
  if (vers > 18) {
    deserialize16(gpsDistToHome);              // 134-135: gpsDistToHome
    deserialize16(gpsDirection);               // 136-137: gpsDirecton
    deserialize8(gpsNumSat);                   // 138: number of satellites
    deserialize8(gpsFix);                      // 139: ?
    deserialize8(gpsUpdate);                   // 140: ?
  }

  deserialize16(powerMeterSum);                // 141-142: powerMeterSum
  deserialize16(powerTrigger);                 // 143-144: powerTrigger
  deserialize8(vbat);                          // 145: vbat (battery level)

  deserialize16(4, debugVals);                 // 146-153: 4 16bit debug values
}

// --------------------------------

void serialWrite() {
  lcdNoBlink();
  
  bptr = buffer;
  int8_t i;
  
  serialize8('W');                             // 00: tag
  if (vers < 20) {
    for (i = 0; i < 7; i++) {                  // 01-15: PID (roll, pitch, yaw, alt, vel), 16-17: pLevel, iLevel, 18: pMag
      serialize8(P8[i]);
      if (i < 6) {
        serialize8(I8[i]); 
        if (i < 5) {
          serialize8(D8[i]);
        }
      }
    }
  }
  else {
    for (i = 0; i < 8; i++) {                  // 01-23: PID (roll, pitch, yaw, alt, vel, gps, lev, mag)
      serialize8(P8[i]);
      serialize8(I8[i]); 
      serialize8(D8[i]);
    }
  }

  serialize8(rcRate8);                         // 24: rcRate (dual rate)
  serialize8(rcExpo8);                         // 25: rcExpo
  serialize8(rollPitchRate);                   // 26: rollPitchRate
  serialize8(yawRate);                         // 27: yawRate
  serialize8(dynThrPID);                       // 28: dyn. Throttle PID

  if (vers < 20) {
    int actLen = (vers == 18)? 6 : 8;  
    for (i = 0; i < actLen; i++) {             // (24-31): activation modes
      serialize8(actMode[i] & 0x3F);
    }
  }
  else {
    for (i = 0; i < AUX_STATES; i++) {         // 29-50: activation modes
      serialize16(actMode[i] & 0x3FFF);
    }
  }

  serialize16(powerTrigger);                   // 51-52: powerTrigger
    
  uint8_t len = bptr - buffer;
  Serial.write(buffer, len);
  serialFlush();

  lcdPrintProg1(writeMsg);
  format(7 + 16, len, 2, 0);

  blinkConfirm();

  serialRead(); // reload the frame
}

// --------------------------------

#ifdef DUMP_BUFFER
void dumpBuffer() {
  // read buffer
  while (serialAvailable() && (bptr - buffer) < BUFFER_SIZE) {
    *bptr++ = serialReadByte();
  }
  
  // print read bytes
  uint8_t avail = (bptr - buffer);
  for (uint8_t i = 0; i < avail; i += 8) {
    lcdClear();
    printHex(0,  i);
    printHex(16, i+4);

    for (uint8_t j = 0; j < 8; j++) {
      if (i + j < avail) {
        uint8_t pos = 4 + j * 3 + 4 * (j > 3);
        uint8_t tmp = buffer[i + j];
        printHex(pos, tmp);
      }
    }
    int8_t key;
    while ((key = getKeyPressed()) < 0);
    waitForKeyUp();

    if (key == KEY_DECR) {
      i -= 16;
    }
  }
}

void printHex(uint8_t pos, uint8_t val) {
  uint8_t hi  = val / 16;
  uint8_t low = val % 16;

  lcdWrite(pos, '0' + hi  + 7 * (hi  > 9));
  lcdWrite(     '0' + low + 7 * (low > 9));
}

#endif // DUMP_BUFFER

